##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import sys

import transaction


failures = 0
calls = 0

nonselect_desc = [('Query', 'STRING', 62, 62, 0, 0, 1),
                  ('Status', 'STRING', 12, 12, 0, 0, 1),
                  ('Calls', 'STRING', 12, 12, 0, 0, 1)]


class QueryError(Exception):
    pass


class DB:

    _p_oid = _p_changed = _registered = None

    defs = {'STRING': 's', 'NUMBER': 'n', 'DATE': 'd'}

    def Database_Connection(self, string):
        # Create a dbi-compatible database connection
        msg = 'attempt to create a database connection for an abstract dbi'
        raise NotImplementedError(msg)

    Database_Error = 'Should be overriden by subclass'

    def __init__(self, connection):
        self.connection = connection
        db = self.db = self.Database_Connection(connection)
        self.cursor = db.cursor()

    def str(self, v, StringType=type('')):
        if v is None:
            return ''
        r = str(v)
        if r[-1:] == 'L' and type(v) is not StringType:
            r = r[:-1]
        return r

    def __inform_commit__(self, *ignored):
        self._registered = None
        self.db.commit()

    def __inform_abort__(self, *ignored):
        self._registered = None
        self.db.rollback()

    def register(self):
        if self._registered:
            return
        transaction.get().register(self)
        self._registered = 1

    def query(self, query_string, max_rows=9999999):
        global failures, calls
        calls = calls + 1
        try:
            c = self.cursor
            self.register()
            queries = filter(None,
                             [x.strip() for x in query_string.split('\0')])
            if not queries:
                raise QueryError('empty query')
            if len(queries) > 1:
                result = []
                for qs in queries:
                    r = c.execute(qs)
                    if r is None:
                        msg = 'select in multiple sql-statement query'
                        raise QueryError(msg)
                    result.append((qs, str(repr(r)), calls))
                desc = nonselect_desc
            else:
                query_string = queries[0]
                r = c.execute(query_string)
                if r is None:
                    result = c.fetchmany(max_rows)
                    desc = c.description
                else:
                    result = ((query_string, str(repr(r)), calls),)
                    desc = nonselect_desc
            failures = 0
            c.close()
        except self.Database_Error as mess:
            c.close()
            self.db.rollback()
            failures = failures + 1
            if (': invalid' in mess and 'PARSE' in mess) or \
               'Error while trying to retrieve text for error' in mess or \
               failures > 100:
                # Either DBI was stupid or we have too many failures.
                # Maybe the db is hosed.  Let's try once to restart it.
                failures = 0
                c.close()
                self.db.close()
                db = self.db = self.Database_Connection(self.connection)
                self.cursor = db.cursor()
                c = self.cursor
                c.execute(query_string)
                result = c.fetchall()
                desc = c.description
            else:
                raise sys.exc_info()

        if result:
            result = '\n'.join(map(lambda row, self=self:
                                   '\t'.join(map(self.str, row)),
                                   result)) + '\n'
        else:
            result = ''

        return (
            '%s\n%s\n%s' % (
                '\t'.join(map(lambda d: d[0], desc)),
                '\t'.join(
                    map(lambda d, defs=self.defs: '%d%s' % (d[2], defs[d[1]]),
                        desc)), result))



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import logging

from six.moves._thread import allocate_lock

from transaction.interfaces import IDataManager
from transaction.interfaces import TransactionFailedError
from zope.interface import implementer

from .TM import TM


LOG = logging.getLogger('Products.ZSQLMethods')
thunk_lock = allocate_lock()


@implementer(IDataManager)
class THUNKED_TM(TM):
    """A big heavy hammer for handling non-thread safe DAs
    """

    def _register(self):
        if not self._registered:
            thunk_lock.acquire()

            try:
                self.transaction_manager.get().join(self)
            except TransactionFailedError:
                LOG.error('Failed to join transaction: ', exc_info=True)
                # No need to raise here, the transaction is already
                # broken as a whole
            except ValueError:
                LOG.error('Failed to join transaction: ', exc_info=True)
                # Raising here, the transaction is in an invalid state
                raise
            else:
                self._begin()
                self._registered = 1
            finally:
                thunk_lock.release()

    def tpc_finish(self, *ignored):
        if self._registered:
            try:
                self._finish()
            finally:
                thunk_lock.release()
                self._registered = 0

    def abort(self, *ignored):
        if self._registered:
            try:
                self._abort()
            finally:
                thunk_lock.release()
                self._registered = 0



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import unittest


class ConnectionTests(unittest.TestCase):

    def _getTargetClass(self):
        from Shared.DC.ZRDB.Connection import Connection
        return Connection

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_connect_on_load(self):

        class _Connection(self._getTargetClass()):

            def connect(self, connection_string):
                setattr(self, '_connected_to', connection_string)

        conn1 = _Connection('conn1', '', 'conn string 1')
        conn1.__setstate__(None)
        self.assertEqual(conn1._connected_to, 'conn string 1')

        conn2 = _Connection('conn2', '', 'conn string 2')
        conn2.connect_on_load = False
        conn2.__setstate__(None)
        self.assertFalse(hasattr(conn2, '_connected_to'))

    def test_sql_quote___miss(self):
        TO_QUOTE = 'no quoting required'
        conn = self._makeOne('conn', '', 'conn string')
        self.assertEqual(conn.sql_quote__(TO_QUOTE), "'%s'" % TO_QUOTE)

    def test_sql_quote___embedded_apostrophe(self):
        TO_QUOTE = "w'embedded apostrophe"
        conn = self._makeOne('conn', '', 'conn string')
        self.assertEqual(conn.sql_quote__(TO_QUOTE),
                         "'w''embedded apostrophe'")

    def test_sql_quote___embedded_backslash(self):
        TO_QUOTE = 'embedded \\backslash'
        conn = self._makeOne('conn', '', 'conn string')
        self.assertEqual(conn.sql_quote__(TO_QUOTE),
                         "'embedded \\backslash'")
        # Show for good measure that the seeming two backslashes
        # are really one, when you look at the raw string.
        self.assertEqual(conn.sql_quote__(TO_QUOTE),
                         r"'embedded \backslash'")

    def test_sql_quote___embedded_double_quote(self):
        # As it turns out, escaping double quotes will break
        # some servers, notably PostgreSQL, so we no longer do that.
        TO_QUOTE = 'embedded "double quote'
        conn = self._makeOne('conn', '', 'conn string')
        self.assertEqual(conn.sql_quote__(TO_QUOTE),
                         "'embedded \"double quote'")

    def test_sql_quote___embedded_null(self):
        TO_QUOTE = "w'embedded apostrophe and \x00null"
        conn = self._makeOne('conn', '', 'conn string')
        self.assertEqual(conn.sql_quote__(TO_QUOTE),
                         "'w''embedded apostrophe and null'")

        # This is another version of a nul character.
        TO_QUOTE = 'embedded other \x1anull'
        conn = self._makeOne('conn', '', 'conn string')
        self.assertEqual(conn.sql_quote__(TO_QUOTE),
                         "'embedded other null'")

    def test_sql_quote___embedded_carriage_return(self):
        TO_QUOTE = "w'embedded carriage\rreturn"
        conn = self._makeOne('conn', '', 'conn string')
        self.assertEqual(conn.sql_quote__(TO_QUOTE),
                         "'w''embedded carriagereturn'")


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(ConnectionTests))
    return suite



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import unittest

from six.moves import UserDict


def _sql_quote(v):
    return '"%s"' % v


class FauxMultiDict(UserDict):

    def getitem(self, key, call):
        if key == 'sql_quote__':
            return _sql_quote

        v = self[key]
        if v is not None:
            if call and callable(v):
                v = v()
        return v


class SQLTestTests(unittest.TestCase):

    def _getTargetClass(self):
        from Shared.DC.ZRDB.sqltest import SQLTest
        return SQLTest

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_string(self):
        tested = self._makeOne('foo type="string"')
        self.assertEqual(tested.render(FauxMultiDict(foo='FOO')),
                         'foo = "FOO"')

    def test_string_binary(self):
        tested = self._makeOne('foo type="string"')
        self.assertEqual(tested.render(FauxMultiDict(foo=b'FOO')),
                         'foo = "FOO"')

    def test_int(self):
        tested = self._makeOne('foo type="int"')

        for valid in (100, 100.00, '100', b'100'):
            self.assertEqual(tested.render(FauxMultiDict(foo=valid)),
                             'foo = 100')

        for invalid in ('', b'', None):
            self.assertRaises(ValueError,
                              tested.render,
                              FauxMultiDict(foo=invalid))

    def test_float(self):
        tested = self._makeOne('foo type="float"')

        for valid in (100, 100.0, '100.0', b'100.0'):
            self.assertEqual(tested.render(FauxMultiDict(foo=valid)),
                             'foo = 100.0')

        for invalid in ('', b'', None):
            self.assertRaises(ValueError,
                              tested.render,
                              FauxMultiDict(foo=invalid))


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(SQLTestTests))
    return suite



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Inserting values with the 'sqlvar' tag

    The 'sqlvar' tag is used to type-safely insert values into SQL
    text.  The 'sqlvar' tag is similar to the 'var' tag, except that
    it replaces text formatting parameters with SQL type information.

    The sqlvar tag has the following attributes:

      name -- The name of the variable to insert. As with other
              DTML tags, the 'name=' prefix may be, and usually is,
              ommitted.

      type -- The data type of the value to be inserted.  This
              attribute is required and may be one of 'string',
              'int', 'float', or 'nb'.  The 'nb' data type indicates a
              string that must have a length that is greater than 0.

      optional -- A flag indicating that a value is optional.  If a
                  value is optional and is not provided (or is blank
                  when a non-blank value is expected), then the string
                  'null' is inserted.

    For example, given the tag::

      <dtml-sqlvar x type=nb optional>

    if the value of 'x' is::

      Let\'s do it

    then the text inserted is:

      'Let''s do it'

    however, if x is ommitted or an empty string, then the value
    inserted is 'null'.
"""

import six

from DocumentTemplate.DT_Util import ParseError
from DocumentTemplate.DT_Util import name_param
from DocumentTemplate.DT_Util import parse_params


StringTypes = six.string_types + (six.binary_type,)


class SQLVar:
    name = 'sqlvar'

    def __init__(self, args, encoding=None):
        self.encoding = encoding
        args = parse_params(args, name='', expr='', type=None, optional=1)

        name, expr = name_param(args, 'sqlvar', 1)
        if expr is None:
            expr = name
        else:
            expr = expr.eval
        self.__name__, self.expr = name, expr

        self.args = args
        if 'type' not in args:
            raise ParseError('the type attribute is required', 'dtvar')
        t = args['type']
        if t not in valid_types:
            raise ParseError('invalid type, %s' % t, 'dtvar')

    def render(self, md):
        name = self.__name__
        args = self.args
        t = args['type']
        try:
            expr = self.expr
            if isinstance(expr, StringTypes):
                v = md[expr]
            else:
                v = expr(md)
        except Exception:
            if 'optional' in args and args['optional']:
                return 'null'
            if not isinstance(expr, StringTypes):
                raise
            raise ValueError('Missing input variable, <em>%s</em>' % name)

        if v is None:
            return 'null'

        if t == 'int':
            try:
                if isinstance(v, StringTypes):
                    if v[-1:] == 'L':
                        v = v[:-1]
                    int(v)
                else:
                    v = str(int(v))
            except Exception:
                if not v and 'optional' in args and args['optional']:
                    return 'null'
                err = 'Invalid integer value for <em>%s</em>' % name
                raise ValueError(err)
        elif t == 'float':
            try:
                if isinstance(v, StringTypes):
                    if v[-1:] == 'L':
                        v = v[:-1]
                    float(v)
                else:
                    v = str(float(v))
            except Exception:
                if not v and 'optional' in args and args['optional']:
                    return 'null'
                err = 'Invalid floating-point value for <em>%s</em>' % name
                raise ValueError(err)
        else:
            if not isinstance(v, (str, StringTypes)):
                v = str(v)
            if not v and t == 'nb':
                if 'optional' in args and args['optional']:
                    return 'null'
                else:
                    err = 'Invalid empty string value for <em>%s</em>' % name
                    raise ValueError(err)

            v = md.getitem('sql_quote__', 0)(v)
            # if v.find("\'") >= 0: v="''".join(v.split("\'"))
            # v="'%s'" % v

        return v

    __call__ = render


valid_types = {'int': 1, 'float': 1, 'string': 1, 'nb': 1}



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Inserting optional tests with 'sqlgroup'

    It is sometimes useful to make inputs to an SQL statement
    optinal.  Doing so can be difficult, because not only must the
    test be inserted conditionally, but SQL boolean operators may or
    may not need to be inserted depending on whether other, possibly
    optional, comparisons have been done.  The 'sqlgroup' tag
    automates the conditional insertion of boolean operators.

    The 'sqlgroup' tag is a block tag that has no attributes. It can
    have any number of 'and' and 'or' continuation tags.

    Suppose we want to find all people with a given first or nick name
    and optionally constrain the search by city and minimum and
    maximum age.  Suppose we want all inputs to be optional.  We can
    use DTML source like the following::

      <dtml-sqlgroup>
        <dtml-sqlgroup>
          <dtml-sqltest name column=nick_name type=nb multiple optional>
        <dtml-or>
          <dtml-sqltest name column=first_name type=nb multiple optional>
        </dtml-sqlgroup>
      <dtml-and>
        <dtml-sqltest home_town type=nb optional>
      <dtml-and>
        <dtml-if minimum_age>
           age >= <dtml-sqlvar minimum_age type=int>
        </dtml-if>
      <dtml-and>
        <dtml-if maximum_age>
           age <= <dtml-sqlvar maximum_age type=int>
        </dtml-if>
      </dtml-sqlgroup>

    This example illustrates how groups can be nested to control
    boolean evaluation order.  It also illustrates that the grouping
    facility can also be used with other DTML tags like 'if' tags.

    The 'sqlgroup' tag checks to see if text to be inserted contains
    other than whitespace characters.  If it does, then it is inserted
    with the appropriate boolean operator, as indicated by use of an
    'and' or 'or' tag, otherwise, no text is inserted.

"""

import six

from DocumentTemplate.DT_Util import ParseError
from DocumentTemplate.DT_Util import name_param
from DocumentTemplate.DT_Util import parse_params


StringTypes = six.string_types + (six.binary_type,)


class SQLTest:
    name = 'sqltest'
    optional = multiple = None

    def __init__(self, args, encoding=None):
        self.encoding = encoding
        args = parse_params(args, name='', expr='', type=None, column=None,
                            multiple=1, optional=1, op=None)
        name, expr = name_param(args, 'sqlvar', 1)

        if expr is None:
            expr = name
        else:
            expr = expr.eval
        self.__name__, self.expr = name, expr

        self.args = args
        if 'type' not in args:
            raise ParseError('the type attribute is required', 'sqltest')

        self.type = t = args['type']
        if t not in valid_types:
            raise ParseError('invalid type, %s' % t, 'sqltest')

        if 'optional' in args:
            self.optional = args['optional']
        if 'multiple' in args:
            self.multiple = args['multiple']
        if 'column' in args:
            self.column = args['column']
        elif self.__name__ is None:
            err = ' the column attribute is required if an expression is used'
            raise ParseError(err, 'sqltest')
        else:
            self.column = self.__name__

        # Deal with optional operator specification
        op = '='                        # Default
        if 'op' in args:
            op = args['op']
            # Try to get it from the chart, otherwise use the one provided
            op = comparison_operators.get(op, op)
        self.op = op

    def render(self, md):
        name = self.__name__

        t = self.type
        args = self.args
        try:
            expr = self.expr
            if isinstance(expr, StringTypes):
                v = md[expr]
            else:
                v = expr(md)
        except KeyError:
            if 'optional' in args and args['optional']:
                return ''
            raise ValueError('Missing input variable, <em>%s</em>' % name)

        if isinstance(v, (list, tuple)):
            if len(v) > 1 and not self.multiple:
                msg = 'multiple values are not allowed for <em>%s</em>' % name
                raise ValueError(msg)
        else:
            v = [v]

        vs = []
        for v in v:
            if not v and isinstance(v, StringTypes) and t != 'string':
                continue
            if t == 'int':
                try:
                    if isinstance(v, StringTypes):
                        if six.PY3 and isinstance(v, bytes):
                            v = v.decode(self.encoding or 'UTF-8')
                        if v[-1:] == 'L':
                            v = v[:-1]
                        int(v)
                    else:
                        v = str(int(v))
                except (TypeError, ValueError):
                    msg = 'Invalid integer value for <em>%s</em>' % name
                    raise ValueError(msg)
            elif t == 'float':
                if not v and isinstance(v, StringTypes):
                    continue
                try:
                    if six.PY3 and isinstance(v, bytes):
                        v = v.decode(self.encoding or 'UTF-8')
                    if isinstance(v, StringTypes):
                        float(v)
                    else:
                        v = str(float(v))
                except (TypeError, ValueError):
                    msg = 'Invalid floating-point value for <em>%s</em>' % name
                    raise ValueError(msg)

            else:
                if not isinstance(v, StringTypes):
                    v = str(v)
                if isinstance(v, six.binary_type):
                    v = v.decode('utf-8')
                # The call to sql_quote__ can return something that is not
                # a native string anymore!
                v = md.getitem('sql_quote__', 0)(v)
                if isinstance(v, six.binary_type):
                    v = v.decode('utf-8')
                # if v.find("\'") >= 0: v="''".(v.split("\'"))
                # v="'%s'" % v
            vs.append(v)

        if not vs and t == 'nb':
            if 'optional' in args and args['optional']:
                return ''
            else:
                err = 'Invalid empty string value for <em>%s</em>' % name
                raise ValueError(err)

        if not vs:
            if self.optional:
                return ''
            raise ValueError('No input was provided for <em>%s</em>' % name)

        if len(vs) > 1:
            vs = ', '.join(map(str, vs))
            if self.op == '<>':
                # Do the equivalent of 'not-equal' for a list,
                # "a not in (b,c)"
                return '%s not in (%s)' % (self.column, vs)
            else:
                # "a in (b,c)"
                return '%s in (%s)' % (self.column, vs)
        return '%s %s %s' % (self.column, self.op, vs[0])

    __call__ = render


valid_types = {'int': 1, 'float': 1, 'string': 1, 'nb': 1}

comparison_operators = {'eq': '=', 'ne': '<>',
                        'lt': '<', 'le': '<=', 'lte': '<=',
                        'gt': '>', 'ge': '>=', 'gte': '>='}



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import ExtensionClass
from Acquisition import Implicit
from Record import Record


class SQLAlias(ExtensionClass.Base):

    def __init__(self, name):
        self._n = name

    def __of__(self, parent):
        return getattr(parent, self._n)


class NoBrains(ExtensionClass.Base):
    pass


class Results:
    """Class for providing a nice interface to DBI result data
    """
    _index = None

    # We need to allow access to not-explicitly-protected
    # individual record objects contained in the result.
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, items_data, brains=NoBrains, parent=None,
                 zbrains=NoBrains):

        items, data = items_data
        self._data = data
        self.__items__ = items
        self._parent = parent
        self._names = names = []
        self._schema = schema = {}
        self._data_dictionary = dd = {}

        aliases = []
        i = 0

        for item in items:

            name = item['name'].strip()

            if not name:
                raise ValueError('Empty column name, %s' % name)

            if name in schema:
                raise ValueError('Duplicate column name, %s' % name)

            schema[name] = i
            i += 1

            n = name.lower()
            if n != name:
                aliases.append((n, SQLAlias(name)))

            n = name.upper()
            if n != name:
                aliases.append((n, SQLAlias(name)))

            dd[name] = item
            names.append(name)

        self._nv = len(names)

        # Create a record class to hold the records.
        names = tuple(names)

        class r(Record, Implicit, brains, zbrains):
            """Result record class"""

        for k, v in Record.__dict__.items():
            if k[:2] == '__':
                setattr(r, k, v)

        r.__record_schema__ = schema

        # Add SQL Aliases
        for k, v in aliases:
            if getattr(r, k, self) is self:
                setattr(r, k, v)

        self._class = r

        # OK, we've read meta data, now get line indexes

    def _searchable_result_columns(self):
        return self.__items__

    def names(self):
        return self._names

    def data_dictionary(self):
        return self._data_dictionary

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        if index == self._index:
            return self._row
        parent = self._parent
        fields = self._class(self._data[index], parent)
        if parent is not None:
            fields = fields.__of__(parent)
        self._index = index
        self._row = fields
        return fields

    def tuples(self):
        return list(map(tuple, self))

    def dictionaries(self):
        return [{name: row[name] for name in self.names()} for row in self]

    def asRDB(self):  # Waaaaa
        r = []
        append = r.append
        strings = []
        nstrings = []
        items = self.__items__
        indexes = range(len(items))

        for i in indexes:
            item = items[i]
            t = item['type'].lower()
            if t == 's' or t == 't':
                t == 't'
                strings.append(i)
            else:
                nstrings.append(i)
            if 'width' in item:
                append('%s%s' % (item['width'], t))
            else:
                r.append(t)

        r = ['\t'.join(self._names), '\t'.join(r)]
        append = r.append
        row = [''] * len(items)
        tostr = str
        for d in self._data:
            for i in strings:
                v = tostr(d[i])
                if v:
                    if v.find('\\') > 0:
                        v = '\\\\'.join(v.split('\\'))
                    if v.find('\t') > 0:
                        v = '\\t'.join(v.split('\t'))
                    if v.find('\n') > 0:
                        v = '\\n'.join(v.split('\n'))
                row[i] = v
            for i in nstrings:
                row[i] = tostr(d[i])
            append('\t'.join(row))
        append('')

        return '\n'.join(r)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""SQL Method Product.
"""

import Shared.DC.ZRDB.Aqueduct
import Shared.DC.ZRDB.RDB
import Shared.DC.ZRDB.Search
import Shared.DC.ZRDB.sqlgroup
import Shared.DC.ZRDB.sqltest
import Shared.DC.ZRDB.sqlvar
from AccessControl.Permissions import open_close_database_connection
from Products.ZSQLMethods import SQL


def initialize(context):

    context.registerClass(
        SQL.SQL,
        permission='Add Database Methods',
        constructors=(SQL.manage_addZSQLMethodForm, SQL.manage_addZSQLMethod),
        icon='sqlmethod.gif',
        permissions=(open_close_database_connection, ),
        legacy=(SQL.SQLConnectionIDs, ))

    context.registerClass(
        meta_type='Z Search Interface',
        permission='Add Documents, Images, and Files',
        constructors=(Shared.DC.ZRDB.Search.addForm,
                      Shared.DC.ZRDB.Search.manage_addZSearch),
        legacy=(Shared.DC.ZRDB.Search.ZQueryIds, ))


__module_aliases__ = (
    ('Products.AqueductSQLMethods', 'Products.ZSQLMethods'),
    ('Aqueduct', Shared.DC.ZRDB),
    ('AqueductDA', Shared.DC.ZRDB),
    ('Products.AqueductSQLMethods.SQL', SQL),
    ('Aqueduct.Aqueduct', Shared.DC.ZRDB.Aqueduct),
    ('AqueductDA.DA', Shared.DC.ZRDB.DA),
    ('Aqueduct.RDB', Shared.DC.ZRDB.RDB),
    ('AqueductDA.sqlvar', Shared.DC.ZRDB.sqlvar),
    ('AqueductDA.sqltest', Shared.DC.ZRDB.sqltest),
    ('AqueductDA.sqlgroup', Shared.DC.ZRDB.sqlgroup),
)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Generic Database Connection Support"""

try:
    from html import escape # noqa
except ImportError:  # Python 2
    from cgi import escape  # noqa
import sys
from logging import getLogger

from six import StringIO

from AccessControl.class_init import InitializeClass
from AccessControl.Permissions import change_database_connections
from AccessControl.Permissions import open_close_database_connection
from AccessControl.Permissions import test_database_connections
from AccessControl.Permissions import view_management_screens
from AccessControl.SecurityInfo import ClassSecurityInfo
from Acquisition import Implicit
from App.Dialogs import MessageDialog
from App.special_dtml import DTMLFile
from DateTime.DateTime import DateTime
from DocumentTemplate import HTML
from DocumentTemplate.DT_Var import sql_quote
from OFS.role import RoleManager
from OFS.SimpleItem import Item
from Persistence import Persistent
from zExceptions import BadRequest

from . import RDB
from .Aqueduct import custom_default_report
from .Results import Results


LOG = getLogger('ZRDB.Connection')


class Connection(Persistent, RoleManager, Item, Implicit):

    security = ClassSecurityInfo()

    # Specify definitions for tabs:
    manage_options = (({'label': 'Status', 'action': 'manage_main'},
                       {'label': 'Properties', 'action': 'manage_properties'},
                       {'label': 'Test', 'action': 'manage_testForm'})
                      + RoleManager.manage_options + Item.manage_options)

    _v_connected = ''
    connection_string = ''
    # Should the database connection be established when the object
    # is loaded from the ZODB (in __setstate__)?
    connect_on_load = True

    def __init__(self, id, title, connection_string, check=None):
        self.id = str(id)
        self.edit(title, connection_string, check)

    def __setstate__(self, state):
        Persistent.__setstate__(self, state)
        if self.connect_on_load and self.connection_string:
            try:
                self.connect(self.connection_string)
            except Exception:
                LOG.error('Error connecting to relational database.',
                          exc_info=True)

    def title_and_id(self):
        s = Connection.inheritedAttribute('title_and_id')(self)
        if hasattr(self, '_v_connected') and self._v_connected:
            s = '%s, which is connected' % s
        else:
            s = '%s, which is <font color=red>not connected</font>' % s
        return s

    def title_or_id(self):
        s = Connection.inheritedAttribute('title_or_id')(self)
        if hasattr(self, '_v_connected') and self._v_connected:
            s = '%s (connected)' % s
        else:
            s = '%s (<font color=red>not connected</font>)' % s
        return s

    def connected(self):
        return self._v_connected

    def edit(self, title, connection_string, check=1):
        self.title = title
        self.connection_string = connection_string
        if check:
            self.connect(connection_string)

    manage_properties = DTMLFile('dtml/connectionEdit', globals())

    @security.protected(change_database_connections)
    def manage_edit(self, title, connection_string, check=None, REQUEST=None):
        """Change connection
        """
        self.edit(title, connection_string, check)
        if REQUEST is not None:
            esc_id = escape(self.id)
            return MessageDialog(
                title='Edited',
                message='<strong>%s</strong> has been edited.' % esc_id,
                action='./manage_main')

    security.declareProtected(test_database_connections,  # NOQA: D001
                              'manage_testForm')
    manage_testForm = DTMLFile('dtml/connectionTestForm', globals())

    @security.protected(test_database_connections)
    def manage_test(self, query, REQUEST=None):
        """Executes the SQL in parameter 'query' and returns results"""
        dbc = self()  # get our connection
        res = dbc.query(query)

        if isinstance(res, type('')):
            f = StringIO()
            f.write(res)
            f.seek(0)
            result = RDB.File(f)
        else:
            result = Results(res)

        if REQUEST is None:
            return result  # return unadulterated result objects

        if result._searchable_result_columns():
            r = custom_default_report(self.id, result)
        else:
            r = 'This statement returned no results.'

        report = HTML(
            '<html><body bgcolor="#ffffff" link="#000099" vlink="#555555">\n'
            '<dtml-var name="manage_tabs">\n<hr>\n%s\n\n'
            '<hr><h4>SQL Used:</strong><br>\n<pre>\n%s\n</pre>\n<hr>\n'
            '</body></html>'
            % (r, query))

        report = report(*(self, REQUEST), **{self.id: result})

        return report

    security.declareProtected(view_management_screens,  # NOQA: D001
                              'manage_main')
    manage_main = DTMLFile('dtml/connectionStatus', globals())

    @security.protected(open_close_database_connection)
    def manage_close_connection(self, REQUEST=None):
        """Close the connection from the ZMI"""
        msg = ''
        try:
            if hasattr(self, '_v_database_connection'):
                self._v_database_connection.close()
        except Exception as exc:
            LOG.error('Error closing relational database connection.',
                      exc_info=True)
            msg = str(exc)
        self._v_connected = ''
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, manage_tabs_message=msg)

    @security.protected(open_close_database_connection)
    def manage_open_connection(self, REQUEST=None):
        """Open the connection from the ZMI"""
        msg = ''
        try:
            self.connect(self.connection_string)
        except Exception as exc:
            LOG.error('Error opening relational database connection.',
                      exc_info=True)
            msg = str(exc)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, manage_tabs_message=msg)

    def __call__(self, v=None):
        try:
            return self._v_database_connection
        except AttributeError:
            s = self.connection_string
            if s:
                self.connect(s)
                return self._v_database_connection
            raise BadRequest('The database connection is not connected')

    def connect(self, s):
        self.manage_close_connection()
        DB = self.factory()
        try:
            try:
                self._v_database_connection = DB(s)
            except Exception:
                t, v, tb = sys.exc_info()
                raise BadRequest(
                    '<strong>Error connecting to DB.</strong><br>\n'
                    '<!--\n%s\n%s\n-->\n' % (t, v)).with_traceback(tb)
        finally:
            tb = None
        self._v_connected = DateTime()

        return self

    def sql_quote__(self, v):
        v = sql_quote(v)
        return "'%s'" % v


InitializeClass(Connection)



# Dummy implementations for unit tests

from Acquisition import Implicit
from ExtensionClass import Base


class DummySQLConnection(Base, Implicit):

    _isAnSQLConnection = True
    meta_type = 'Dummy ZSQLConnection'

    def __init__(self, id, title=''):
        self._id = id
        self.title = title

    def id(self):
        return self._id

    def title_and_id(self):
        if not self.title:
            return self._id
        return '%s (%s)' % (self.title, self._id)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import binascii
import os
import re

from six import StringIO

from Acquisition import Implicit
from App.Common import package_home
from DateTime.DateTime import DateTime
from DocumentTemplate import HTML
from DocumentTemplate import File
from OFS.role import RoleManager
from OFS.SimpleItem import Item
from Persistence import Persistent
from zExceptions import Redirect


dtml_dir = os.path.join(package_home(globals()), 'dtml')
InvalidParameter = 'Invalid Parameter'


class BaseQuery(Persistent, Item, Implicit, RoleManager):

    _col = None
    _arg = {}
    query_date = DateTime()
    manage_options = ()
    MissingArgumentError = 'Bad Request'

    def query_year(self):
        return self.query_date.year()

    def query_month(self):
        return self.query_date.month()

    def query_day(self):
        return self.query_date.day()

    def quoted_input(self):
        return quotedHTML(self.input_src)

    def quoted_report(self):
        return quotedHTML(self.report_src)

    def _convert(self):
        self._arg = parse(self.arguments_src)

    def _argdata(self, REQUEST):
        r = {}

        try:
            args = self._arg
        except Exception:
            self._convert()
            args = self._arg

        id = self.id
        missing = []

        for name in args.keys():
            idname = '%s/%s' % (id, name)
            try:
                r[name] = REQUEST[idname]
            except Exception:
                try:
                    r[name] = REQUEST[name]
                except Exception:
                    arg = args[name]
                    try:
                        r[name] = arg['default']
                    except Exception:
                        try:
                            if not arg['optional']:
                                missing.append(name)
                        except Exception:
                            missing.append(name)

        # Note: the code above tries to check if an argument of the
        # ZSQL method above has the "optional" flag set (in case the
        # argument is omitted from the ZSQL function call). But there
        # is neither corresponding code inside the parse() function to
        # check for the "optional" parameter nor any documentation.
        # So we omit the check for the optional parameter. There will
        # be probably no code break but there will be hopefully more code
        # to work as supposed to work.

#        if missing:
#            raise self.MissingArgumentError(  \
#                "The following arguments were omitted " \
#                " from the ZSQL method call: %s" % str(missing))
#

        return r


class Searchable(BaseQuery):

    def _searchable_arguments(self):

        try:
            return self._arg
        except Exception:
            self._convert()
            return self._arg

    def _searchable_result_columns(self):
        return self._col

    def manage_testForm(self, REQUEST):
        """Provide testing interface"""
        input_src = default_input_form(self.title_or_id(),
                                       self._searchable_arguments(),
                                       'manage_test')
        return HTML(input_src)(self, REQUEST)

    def manage_test(self, REQUEST):
        """Perform an actual query"""

        result = self(REQUEST)
        report = HTML(custom_default_report(self.id, result))
        return report(*(self, REQUEST), **{self.id: result})

    def index_html(self, URL1):
        """ """
        raise Redirect('%s/manage_testForm' % URL1)


class Composite(object):

    def _getquery(self, id):

        o = self
        i = 0
        while 1:
            __traceback_info__ = o
            q = getattr(o, id)
            try:
                if hasattr(q, '_searchable_arguments'):
                    try:
                        q = q.__of__(self.aq_parent)
                    except Exception:
                        pass
                    return q
            except Exception:
                pass
            if i > 100:
                raise AttributeError(id)
            i = i + 1
            o = o.aq_parent

    def myQueryIds(self):
        return map(
            lambda k, queries=self.queries:
            {'id': k, 'selected': k in queries},
            self.ZQueryIds())


def default_input_form(id, arguments, action='query', tabs=''):
    if arguments:
        items = arguments.items()
        return (
            '%s\n%s%s' % (
                '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN"'
                ' "http://www.w3.org/TR/REC-html40/loose.dtd">\n'
                '<html lang="en"><head><title>%s Input Data</title></head>\n'
                '<body bgcolor="#FFFFFF" link="#000099" vlink="#555555">\n%s\n'
                '<form action="&dtml-URL2;/&dtml-id;/%s" '
                'method="get">\n'
                '<h2>%s Input Data</h2>\n'
                'Enter query parameters:<br>'
                '<table>\n'
                % (id, tabs, action, id),
                '/n'.join(
                    map(
                        lambda a:
                        ('<tr> <th>%s</th>\n'
                         '     <td><input name="%s"\n'
                         '                size="30" value="%s">'
                         '     </td></tr>'
                         % (nicify(a[0]),
                            'type' in a[1]
                             and ('%s:%s' % (a[0], a[1]['type']))
                             or a[0],
                            'default' in a[1] and a[1]['default'] or '')),
                            items)),
                '\n<tr><td colspan=2 align=center>\n'
                '<input type="SUBMIT" name="SUBMIT" value="Submit Query">\n'
                '<dtml-if HTTP_REFERER>\n'
                '  <input type="SUBMIT" name="SUBMIT" value="Cancel">\n'
                '  <INPUT NAME="CANCEL_ACTION" TYPE="HIDDEN"\n'
                '         VALUE="&dtml-HTTP_REFERER;">\n'
                '</dtml-if>\n'
                '</td></tr>\n</table>\n</form>\n</body>\n</html>\n'))
    else:
        return (
            '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN" '
            '"http://www.w3.org/TR/REC-html40/loose.dtd">\n'
            '<html lang="en"><head><title>%s Input Data</title></head>\n'
            '<body bgcolor="#FFFFFF" link="#000099" vlink="#555555">\n%s\n'
            '<form action="&dtml-URL2;/&dtml-id;/%s" '
            'method="get">\n'
            '<h2>%s Input Data</h2>\n'
            'This query requires no input.<p>\n'
            '<input type="SUBMIT" name="SUBMIT" value="Submit Query">\n'
            '<dtml-if HTTP_REFERER>\n'
            '  <input type="SUBMIT" name="SUBMIT" value="Cancel">\n'
            '  <INPUT NAME="CANCEL_ACTION" TYPE="HIDDEN"\n'
            '         VALUE="&dtml-HTTP_REFERER;">\n'
            '</dtml-if>\n'
            '</td></tr>\n</table>\n</form>\n</body>\n</html>\n'
            % (id, tabs, action, id))


custom_default_report_src = File(
    os.path.join(dtml_dir, 'customDefaultReport.dtml'))
custom_default_zpt_report_src = File(
    os.path.join(dtml_dir, 'customDefaultZPTReport.dtml'))


def custom_default_report(id, result, action='', no_table=0,
                          goofy=re.compile(r'\W').search):

    columns = result._searchable_result_columns()
    __traceback_info__ = columns
    heading = ('<tr>\n%s        </tr>' %
               ''.join(
                   map(lambda c:
                       '          <th>%s</th>\n' % nicify(c['name']),
                       columns)))

    if no_table:
        tr, _tr, td, _td, delim = '<p>', '</p>', '', '', ',\n'
    else:
        tr, _tr, td, _td, delim = '<tr>', '</tr>', '<td>', '</td>', '\n'

    row = []
    for c in columns:
        n = c['name']
        if goofy(n) is not None:
            n = 'expr="_[\'%s]"' % (repr('"' + n)[2:])
        row.append('          %s<dtml-var %s%s>%s'
                   % (td, n, c['type'] != 's' and ' null=""' or '', _td))

    row = ('     %s\n%s\n        %s' % (
        tr, delim.join(row), _tr))

    return custom_default_report_src(
        id=id, heading=heading, row=row, action=action, no_table=no_table)


def custom_default_zpt_report(id, result, action='', no_table=0,
                              goofy=re.compile(r'\W').search):

    columns = result._searchable_result_columns()
    __traceback_info__ = columns
    heading = ('<tr>\n%s        </tr>' %
               ''.join(
                   map(lambda c:
                       '          <th>%s</th>\n' % nicify(c['name']),
                       columns)))

    if no_table:
        tr, _tr, td, _td, delim = '<p>', '</p>', '', '', ',\n'
    else:
        tr, _tr, td, _td, delim = '<tr>', '</tr>', '<td>', '</td>', '\n'

    row = []
    for c in columns:
        n = c['name']
        tpl = '          %s<span tal:replace="result/%s">%s goes here</span>%s'
        row.append(tpl % (td, n, n, _td))

    row = ('     %s\n%s\n        %s' % (
        tr, delim.join(row), _tr))

    return custom_default_zpt_report_src(
        id=id, heading=heading, row=row, action=action, no_table=no_table)


def detypify(arg):
    idx = arg.find(':')
    if idx > 0:
        arg = arg[:idx]
    return arg


def decode(input, output):
    while 1:
        line = input.readline()
        if not line:
            break
        s = binascii.a2b_base64(line[:-1])
        output.write(s)


def decodestring(s):
    f = StringIO(s)
    g = StringIO()
    decode(f, g)
    return g.getvalue()


class Args(object):

    def __init__(self, data=None, keys=None):
        self._data = data or {}
        self._keys = keys or []

    def items(self):
        return map(lambda k, d=self._data: (k, d[k]), self._keys)

    def values(self):
        return map(lambda k, d=self._data: d[k], self._keys)

    def keys(self):
        return list(self._keys)

    def has_key(self, key):
        return key in self._data

    def __contains__(self, key):
        return key in self._data

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, v):
        self._data[key] = v

    def __delitem__(self, key):
        del self._data[key]

    def __len__(self):
        return len(self._data)


def parse(text,
          result=None,
          keys=None,
          unparmre=re.compile(
              r'([\000- ]*([^\000- ="]+))'),
          parmre=re.compile(
              r'([\000- ]*([^\000- ="]+)=([^\000- ="]+))'),
          qparmre=re.compile(
              r'([\000- ]*([^\000- ="]+)="([^"]*)")'),
          ):

    if result is None:
        result = {}
        keys = []

    __traceback_info__ = text

    mo = parmre.match(text)

    if mo:
        name = mo.group(2)
        value = {'default': mo.group(3)}
        group_len = len(mo.group(1))

    else:
        mo = qparmre.match(text)

        if mo:
            name = mo.group(2)
            value = {'default': mo.group(3)}
            group_len = len(mo.group(1))

        else:
            mo = unparmre.match(text)

            if mo:
                name = mo.group(2)
                value = {}
                group_len = len(mo.group(1))
            else:
                if not text or not text.strip():
                    return Args(result, keys)
                raise InvalidParameter(text)

    lt = name.find(':')
    if lt > 0:
        value['type'] = name[lt + 1:]
        name = name[:lt]

    result[name] = value
    keys.append(name)

    return parse(text[group_len:], result, keys)


def quotedHTML(text,
               character_entities=(
                   ('&', '&amp;'),
                   ('<', '&lt;'),
                   ('>', '&gt;'),
                   ('"', '&quot;'))):

    for char, name in character_entities:
        text = text.replace(char, name)

    return text


def nicify(name):
    name = name.strip().replace('_', ' ')
    return name[:1].upper() + name[1:]


def decapitate(html, RESPONSE=None,
               header_re=re.compile(
                   r'(('
                   r'[^\000- <>:]+:[^\n]*\n'
                   r'|'
                   r'[ \011]+[^\000- ][^\n]*\n'
                   r')+)[ \t]*\n([\000-\377]+)'),  # please kill me now
               space_re=re.compile(r'([ \t]+)'),
               name_re=re.compile(r'([^\000- <>:]+):([^\n]*)'),
               ):

    mo = header_re.match(html)
    if mo is None:
        return html

    headers, html = mo.group(1, 3)

    headers = headers.split('\n')

    i = 1
    while i < len(headers):
        if not headers[i]:
            del headers[i]
        else:
            mo = space_re.match(headers[i])
            if mo:
                headers[i - 1] = '%s %s' % (
                    headers[i - 1], headers[i][len(mo.group(1)):])
                del headers[i]
            else:
                i = i + 1

    for i in range(len(headers)):
        mo = name_re.match(headers[i])
        if mo:
            k, v = mo.group(1, 2)
            v = v.strip()
        else:
            raise ValueError('Invalid Header (%d): %s ' % (i, headers[i]))
        RESPONSE.setHeader(k, v)

    return html


def delimited_output(results, REQUEST, RESPONSE):
    delim = REQUEST['output-delimiter']
    try:
        output_type = REQUEST['output-type']
    except Exception:
        output_type = 'text/plain'

    RESPONSE.setHeader('content-type', output_type)
    return '%s\n%s\n' % (
        delim.join(results.names()),
        '\n'.join(map(lambda row, delim=delim:
                  delim.join(map(str, row)),
                  results)),
    )



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Provide support for linking an external transaction manager with Zope's
"""
import logging
from warnings import warn

import transaction
from transaction.interfaces import IDataManager
from transaction.interfaces import TransactionFailedError
from zope.interface import implementer


LOG = logging.getLogger('Products.ZSQLMethods')


@implementer(IDataManager)
class TM:
    """Mix-in class that provides transaction management support

    A sub class should call self._register() whenever it performs any
    transaction-dependent operations (e.g. sql statements).

    The sub class will need to override _finish, to finalize work,
    _abort, to roll-back work, and perhaps _begin, if any work is
    needed at the start of a transaction.

    A subclass that uses locking during transaction commit must
    define a sortKey() method.
    """

    _registered = None
    transaction_manager = transaction.manager

    def _begin(self):
        pass

    def _register(self):
        if not self._registered:
            try:
                self.transaction_manager.get().join(self)
            except TransactionFailedError:
                LOG.error('Failed to join transaction: ', exc_info=True)
                # No need to raise here, the transaction is already
                # broken as a whole
            except ValueError:
                LOG.error('Failed to join transaction: ', exc_info=True)
                # Raising here, the transaction is in an invalid state
                raise
            else:
                self._begin()
                self._registered = 1
                self._finalize = 0

    def tpc_begin(self, *ignored):
        pass

    commit = tpc_begin

    def _finish(self):
        self.db.commit()

    def _abort(self):
        self.db.rollback()

    def tpc_vote(self, *ignored):
        self._finalize = 1

    def tpc_finish(self, *ignored):

        if self._finalize:
            try:
                self._finish()
            finally:
                self._registered = 0

    __inform_commit__ = tpc_finish

    def abort(self, *ignored):
        try:
            self._abort()
        finally:
            self._registered = 0

    tpc_abort = abort
    __inform_abort__ = abort

    # Most DA's talking to RDBMS systems do not care about commit order, so
    # return a constant. Must be a string according to ITransactionManager.
    _sort_key = '1'

    def sortKey(self, *ignored):
        """ The sortKey method is used by the transaction subsystem to have a
            known commit order for lock acquisition.
        """
        return self._sort_key

    def setSortKey(self, sort_key):
        self._sort_key = str(sort_key)


# BBB version 4
class Surrogate:

    def __init__(self, db):
        warn(u'The Surrogate class is deprecated and will be removed in '
             u'Products.ZSQLMethods version 4.')
        self._p_jar = db
        self.__inform_commit__ = db.tpc_finish
        self.__inform_abort__ = db.tpc_abort



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from unittest import TestCase
from unittest import TestSuite
from unittest import makeSuite


class TestTM(TestCase):

    def _getTargetClass(self):
        from ..TM import TM
        return TM

    def _makeOne(self):
        return self._getTargetClass()()

    def test_conformance(self):
        from transaction.interfaces import IDataManager
        from zope.interface.verify import verifyObject
        tm = self._makeOne()
        verifyObject(IDataManager, tm)

    def test__register(self):
        tm = self._makeOne()
        self.assertFalse(tm._registered)

        tm._register()
        self.assertTrue(tm._registered)

        # Registering again won't break anything
        tm._register()
        self.assertTrue(tm._registered)

    def test_sortKey(self):
        tm = self._makeOne()
        # the default Transaction Manager should have .sortKey() of '1' for
        # backward compatibility. It must be a string according to the
        # ITransactionManager interface.
        self.assertEqual(tm.sortKey(), '1')

        # but the sortKey() should be adjustable
        tm.setSortKey('2')
        self.assertEqual(tm.sortKey(), '2')

        tm.setSortKey([])
        self.assertEqual(tm.sortKey(), '[]')


def test_suite():
    return TestSuite((makeSuite(TestTM),))



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Search Interface Wizard"""

try:
    from html import escape  # noqa
except ImportError:
    from cgi import escape  # noqa

from AccessControl import getSecurityManager
from App.special_dtml import DTMLFile
from zExceptions import Unauthorized

from .Aqueduct import Args
from .Aqueduct import custom_default_report
from .Aqueduct import custom_default_zpt_report
from .Aqueduct import nicify


addForm = DTMLFile('dtml/searchAdd', globals())


def manage_addZSearch(self, report_id, report_title, report_style,
                      input_id, input_title, object_type, queries=[],
                      REQUEST=None):
    """add a report"""

    if not queries:
        raise ValueError('No <em>searchable objects</em> were selected')

    if not report_id:
        raise ValueError('No <em>report id</em> were specified')

    if input_title and not input_id:
        raise ValueError('No <em>input id</em> were specified')

    qs = map(lambda q, self=self: _getquery(self, q), queries)
    arguments = {}
    keys = []

    checkPermission = getSecurityManager().checkPermission

    for q in qs:
        url = q.absolute_url()
        if input_id:
            for name, arg in q._searchable_arguments().items():
                if len(qs) > 1:
                    key = '%s/%s' % (id, name)
                else:
                    key = name
                arguments[key] = arg
                keys.append(key)
        if q._searchable_result_columns() is None:
            raise ValueError(
                """The input searchable object, <em>%s</em>,
                has not been tested.  Until it has been tested,
                it\'s output schema is unknown, and a report
                cannot be generated.  Before creating a report
                from this query, you must try out the query.  To
                try out the query, <a href="%s">click here</a>.
                """ % (escape(q.title_and_id()), escape(url, 1)))

    if object_type == 'dtml_methods':

        if not checkPermission('Add DTML Methods', self):
            raise Unauthorized('You are not authorized to add DTML Methods.')

        if input_id:
            arguments = Args(arguments, keys)
            self.manage_addDocument(
                input_id, input_title,
                default_input_form(arguments, report_id))

        def _report(obj, style):
            return custom_default_report(obj.id, obj, no_table=style)

        self.manage_addDocument(
            report_id, report_title,
            ('<html><head><title><dtml-var title_or_id></title>'
             '</head><body bgcolor="#FFFFFF">\n%s\n'
             '</body></html>' %
             '\n<hr>\n'.join(map(_report(q, report_style), qs))))

        if REQUEST:
            return self.manage_main(self, REQUEST)

    elif object_type == 'page_templates':

        if not checkPermission('Add Page Templates', self):
            raise Unauthorized('You are not authorized to add Page Templates.')

        if input_id:
            arguments = Args(arguments, keys)
            self.manage_addProduct['PageTemplates'].manage_addPageTemplate(
                input_id, input_title,
                default_input_zpt_form(arguments, report_id))

        def _report(obj, style):
            return custom_default_zpt_report(obj.id, obj, no_table=style)

        self.manage_addProduct['PageTemplates'].manage_addPageTemplate(
            report_id, report_title,
            ('<html><body>\n%s\n'
             '</body></html>' %
             '\n<hr>\n'.join(map(_report(q, report_style), qs))))

        if REQUEST:
            return self.manage_main(self, REQUEST)


def ZQueryIds(self):
    # Note that report server configurations will expend on this
    t = []
    ids = {}
    old = ids.__contains__
    o = self
    n = 0
    while 1:

        # Look for queries
        try:
            map = o.objectMap()
        except AttributeError:
            map = ()

        for i in map:
            try:
                id = i['id']
                if not old(id) and \
                   hasattr(getattr(o, id), '_searchable_arguments'):
                    t.append(i['id'])
                    ids[id] = 1
            except Exception:
                pass

        # Now extend search to parent
        try:
            o = o.aq_parent
        except Exception:
            return t

        if n > 100:
            return t  # Seat belt

        n = n + 1


def _getquery(self, id):

    o = self
    i = 0
    while 1:
        __traceback_info__ = o
        q = getattr(o, id)
        try:
            if hasattr(q, '_searchable_arguments'):
                try:
                    q = q.__of__(self.aq_parent)
                except Exception:
                    pass
                return q
        except Exception:
            pass
        if i > 100:
            raise AttributeError(id)
        i = i + 1
        o = o.aq_parent


def default_input_form(arguments, action='query',
                       tabs=''):
    if arguments:
        items = arguments.items()
        return (
            '%s\n%s%s' % (
                '<html><head><title><dtml-var title_or_id></title>'
                '</head><body bgcolor="#FFFFFF">\n%s\n'
                '<form action="%s" method="get">\n'
                '<h2><dtml-var document_title></h2>\n'
                'Enter query parameters:<br>'
                '<table>\n'
                % (tabs, action),
                '\n'.join(
                    map(
                        lambda a:
                        ('<tr><th>%s</th>\n'
                         '    <td><input name="%s"\n'
                         '               size="30" value="%s">'
                         '</td></tr>'
                         % (nicify(a[0]),
                            'type' in a[1]
                            and ('%s:%s' % (a[0], a[1]['type'])) or a[0],
                            'default' in a[1] and a[1]['default'] or '',
                            )), items)),
                '\n<tr><td colspan=2 align=center>\n'
                '<input type="SUBMIT" name="SUBMIT" value="Submit Query">\n'
                '</td></tr>\n</table>\n</form>\n'
                '</body></html>\n'))
    else:
        return (
            '<html><head><title><dtml-var title_or_id></title>'
            '</head><body bgcolor="#FFFFFF">\n%s\n'
            '<form action="%s" method="get">\n'
            '<h2><dtml-var document_title></h2>\n'
            'This query requires no input.<p>\n'
            '<input type="SUBMIT" name="SUBMIT" value="Submit Query">\n'
            '</form>\n'
            '</body></html>\n' % (tabs, action))


def default_input_zpt_form(arguments, action='query', tabs=''):
    if arguments:
        items = arguments.items()
        return (
            '%s\n%s%s' % (
                '<html><body>\n%s\n'
                '<form action="%s" method="get">\n'
                '<h2 tal:content="template/title_or_id">Title</h2>\n'
                'Enter query parameters:<br>'
                '<table>\n'
                % (tabs, action),
                '\n'.join(
                    map(
                        lambda a:
                        ('<tr><th>%s</th>\n'
                         '    <td><input name="%s"\n'
                         '               size="30" value="%s">'
                         '</td></tr>'
                         % (nicify(a[0]),
                            'type' in a[1]
                            and ('%s:%s' % (a[0], a[1]['type'])) or a[0],
                            'default' in a[1] and a[1]['default'] or '',
                            )), items)),
                '\n<tr><td colspan=2 align=center>\n'
                '<input type="SUBMIT" name="SUBMIT" value="Submit Query">\n'
                '</td></tr>\n</table>\n</form>\n'
                '</body></html>\n'))
    else:
        return (
            '<html><body>\n%s\n'
            '<form action="%s" method="get">\n'
            '<h2 tal:content="template/title_or_id">Title</h2>\n'
            '<p>This query requires no input.</p>\n'
            '<input type="SUBMIT" name="SUBMIT" value="Submit Query">\n'
            '</form>\n'
            '</body></html>\n' % (tabs, action))



import unittest


class SQLMethodTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.ZSQLMethods.SQL import SQL
        return SQL

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_class_conforms_to_IWriteLock(self):
        from zope.interface.verify import verifyClass
        try:
            from OFS.interfaces import IWriteLock
        except ImportError:
            from webdav.interfaces import IWriteLock
        verifyClass(IWriteLock, self._getTargetClass())


class SQLConnectionIdsTests(unittest.TestCase):

    def setUp(self):
        from OFS.Folder import Folder
        from Products.ZSQLMethods.tests.dummy import DummySQLConnection
        super(SQLConnectionIdsTests, self).setUp()

        self.root = Folder('root')
        conn1 = DummySQLConnection('conn1', 'Title1')
        self.root._setObject('conn1', conn1)
        self.root._setObject('child1', Folder('child1'))
        conn2 = DummySQLConnection('conn2', 'Title2')
        self.root.child1._setObject('conn2', conn2)
        self.root._setObject('child2', Folder('child2'))
        conn3 = DummySQLConnection('conn3')
        self.root.child2._setObject('conn3', conn3)
        self.root.child1._setObject('grandchild1', Folder('grandchild1'))
        conn4 = DummySQLConnection('conn4')
        self.root.child1.grandchild1._setObject('conn4', conn4)

    def test_SQLConnectionIDs(self):
        from Products.ZSQLMethods.SQL import SQLConnectionIDs

        self.assertEqual(SQLConnectionIDs(self.root),
                         [('Title1 (conn1)', 'conn1')])
        self.assertEqual(SQLConnectionIDs(self.root.child1),
                         [('Title1 (conn1)', 'conn1'),
                          ('Title2 (conn2)', 'conn2')])
        self.assertEqual(SQLConnectionIDs(self.root.child1.grandchild1),
                         [('Title1 (conn1)', 'conn1'),
                          ('Title2 (conn2)', 'conn2'),
                          ('conn4', 'conn4')])
        self.assertEqual(SQLConnectionIDs(self.root.child2),
                         [('Title1 (conn1)', 'conn1'),
                          ('conn3', 'conn3')])


def test_suite():
    return unittest.TestSuite((unittest.makeSuite(SQLMethodTests),
                               unittest.makeSuite(SQLConnectionIdsTests)))



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Class for reading RDB files"""

import array
import re

import six

from Acquisition import Implicit
from DateTime import DateTime
from ExtensionClass import Base
from Missing import MV
from Record import Record


def parse_text(s):
    if s.find('\\') < 0 and s.find('\\t') < 0 and s.find('\\n') < 0:
        return s
    r = []
    for x in s.split('\\\\'):
        x = '\n'.join(x.split('\\n'))
        r.append('\t'.join(x.split('\\t')))
    return '\\'.join(r)


if six.PY3:
    long = int

Parsers = {'n': float,
           'i': int,
           'l': long,
           'd': DateTime,
           't': parse_text}


class SQLAlias(Base):

    def __init__(self, name):
        self._n = name

    def __of__(self, parent):
        return getattr(parent, self._n)


class NoBrains(Base):
    pass


class DatabaseResults(object):
    """Class for reading RDB files
    """
    _index = None

    # We need to allow access to not-explicitly-protected
    # individual record objects contained in the result.
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, file, brains=NoBrains, parent=None, zbrains=NoBrains):

        self._file = file
        readline = file.readline
        line = readline()
        self._parent = parent

        while line and line.find('#') != -1:
            line = readline()

        line = line[:-1]
        if line and line[-1:] in '\r\n':
            line = line[:-1]
        self._names = names = [name.strip() for name in line.split('\t')]
        if not names:
            raise ValueError('No column names')

        aliases = []
        self._schema = schema = {}
        i = 0
        for name in names:
            if not name:
                raise ValueError('Empty column name, %s' % name)

            if name in schema:
                raise ValueError('Duplicate column name, %s' % name)

            schema[name] = i
            i = i + 1

            n = name.lower()
            if n != name:
                aliases.append((n, SQLAlias(name)))

            n = name.upper()
            if n != name:
                aliases.append((n, SQLAlias(name)))

        self._nv = nv = len(names)
        line = readline()
        line = line[:-1]
        if line[-1:] in '\r\n':
            line = line[:-1]

        self._defs = defs = [_def.strip() for _def in line.split('\t')]

        if not defs:
            raise ValueError('No column definitions')

        if len(defs) != nv:
            raise ValueError(
                """The number of column names and the number of column
                definitions are different.""")

        i = 0
        self._parsers = parsers = []
        defre = re.compile(r'([0-9]*)([a-zA-Z])?')
        self._data_dictionary = dd = {}
        self.__items__ = items = []

        for _def in defs:
            if not _def:
                raise ValueError('Empty column definition for %s' % names[i])

            mo = defre.match(_def)
            if mo is None:
                err = 'Invalid column definition for, %s, for %s' % (
                      _def, names[i])
                raise ValueError(err)

            type = mo.group(2).lower()
            width = mo.group(1)

            if width:
                width = int(width)
            else:
                width = 8

            parser = Parsers.get(type, str)

            name = names[i]
            d = {'name': name, 'type': type, 'width': width, 'parser': parser}
            items.append(d)
            dd[name] = d

            parsers.append((i, parser))
            i += 1

        # Create a record class to hold the records.
        names = tuple(names)

        class r(Record, Implicit, brains, zbrains):
            """Result record class"""

        for k, v in Record.__dict__.items():
            if k[:2] == '__':
                setattr(r, k, v)

        r.__record_schema__ = schema

        # Add SQL Aliases
        for k, v in aliases:
            if getattr(r, k, self) is self:
                setattr(r, k, v)

        self._class = r

        # OK, we've read meta data, now get line indexes

        pos = file.tell()
        save = self._lines = array.array('i')
        save = save.append
        line = readline()
        while line:
            save(pos)
            pos += len(line)
            line = readline()

    def _searchable_result_columns(self):
        return self.__items__

    def names(self):
        return self._names

    def data_dictionary(self):
        return self._data_dictionary

    def __len__(self):
        return len(self._lines)

    def __getitem__(self, index):
        if index == self._index:
            return self._row
        file = self._file
        file.seek(self._lines[index])
        line = file.readline()
        line = line[:-1]
        if line and line[-1:] in '\r\n':
            line = line[:-1]
        fields = [field.strip() for field in line.split('\t')]
        fields_len = len(fields)
        nv = self._nv
        if fields_len != nv:
            if fields_len < nv:
                fields = fields + [''] * (nv - fields_len)
            else:
                raise ValueError(
                    """The number of items in record %s is invalid
                    <pre>%s\n%s\n%s\n%s</pre>
                    """
                    % (index, ('=' * 40), line, ('=' * 40), fields))
        for i, parser in self._parsers:
            try:
                v = parser(fields[i])
            except Exception:
                if fields[i]:
                    raise ValueError(
                        """Invalid value, %s, for %s in record %s"""
                        % (fields[i], self._names[i], index))
                else:
                    v = MV
            fields[i] = v

        parent = self._parent
        record = self._class(fields, parent)
        self._index = index
        self._row = record
        if parent is None:
            return record
        return record.__of__(parent)


File = DatabaseResults



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import pickle
from unittest import TestCase
from unittest import TestSuite
from unittest import makeSuite


class TestArgs(TestCase):

    def _getTargetClass(self):
        from Shared.DC.ZRDB.Aqueduct import Args
        return Args

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_unpickle_old(self):
        # ZSQL Methods created under Zope 2 used an Args class that did not
        # subclass ``object``. Their pickled instances do not load under
        # the newer Args class without code changes.
        old_pickle = (b"(iShared.DC.ZRDB.Aqueduct\nArgs\np0\n(dp1\nS'_data'\n"
                      b"p2\n(dp3\nS'arg1'\np4\n(dp5\nS'default'\np6\nS'n/a'\n"
                      b"p7\nsS'type'\np8\nS'string'\np9\nsssS'_keys'\np10\n"
                      b'(lp11\ng4\nasb.')
        args = pickle.loads(old_pickle)
        self.assertIn('arg1', args)
        self.assertDictEqual(args['arg1'],
                             {'default': 'n/a', 'type': 'string'})

    def test_args_missing(self):
        args = self._makeOne()
        self.assertEqual(len(args), 0)
        self.assertEqual(list(args.items()), [])

    def test_args_invalid(self):
        args = self._makeOne('foo')
        self.assertEqual(list(args.items()), [])

    def test_args_correct(self):
        args = self._makeOne({'arg1': {'type': 'string'}}, ['arg1'])
        self.assertEqual(len(args), 1)

    def test_dict_getters(self):
        args = self._makeOne({'arg1': {'type': 'string', 'default': 'n/a'}},
                             ['arg1'])
        self.assertIn('arg1', args)
        self.assertTrue(args.has_key('arg1'))  # NOQA: flake8: W601
        self.assertDictEqual(args['arg1'],
                             {'default': 'n/a', 'type': 'string'})
        self.assertEqual(args.keys(), ['arg1'])
        self.assertEqual(list(args.values()),
                         [{'type': 'string', 'default': 'n/a'}])
        self.assertEqual(list(args.items()),
                         [('arg1', {'default': 'n/a', 'type': 'string'})])

    def test__setitem__(self):
        args = self._makeOne()
        args['arg2'] = {'type': 'int'}
        self.assertEqual(len(args), 1)
        self.assertIn('arg2', args)
        self.assertDictEqual(args['arg2'], {'type': 'int'})

    def test__delitem__(self):
        args = self._makeOne()
        args['arg2'] = {'type': 'int'}
        self.assertEqual(len(args), 1)

        self.assertRaises(KeyError, args.__delitem__, 'unknown')

        del args['arg2']
        self.assertEqual(len(args), 0)


def test_suite():
    return TestSuite((makeSuite(TestArgs),))



from pprint import pprint
from time import time
from unittest import TestCase
from unittest import TestSuite
from unittest import makeSuite


class DummyDB:

    conn_num = 1
    result = None

    def query(self, query, max_rows):
        if self.result:
            return self.result
        return 'result for ' + query

    def hook_method(self):
        conn_to_use = 'conn' + str(self.conn_num)
        self.conn_num += 1
        return conn_to_use


class DummyTime:

    def __init__(self, t):
        self.t = float(t)

    def __call__(self):
        return self.t


class TestCaching(TestCase):

    echo = False

    def setUp(self):
        from Shared.DC.ZRDB import DA
        self.DA = DA
        self.da = DA.DA('da', 'title', 'conn_id', 'arg1 arg2', 'some sql')
        # set the da's caching parameters
        self.da.cache_time_ = 10
        self.da.max_cache_ = 2

    def _do_query(self, query, t):
        try:
            self.DA.time = DummyTime(t)
            result = self.da._cached_result(DummyDB(), query, 1, 'conn_id')
        finally:
            self.DA.time = time
        self.assertEqual(result, 'result for ' + query)

    def _check_mapping(self, expected, actual):
        missing = []
        extra = []
        different = []
        for key, value in expected.items():
            try:
                ai = actual[key]
            except KeyError:
                missing.append(key)
            else:
                if ai != value:
                    different.append('%r: %r != %r' % (key, value, ai))
        for key in actual.keys():
            try:
                expected[key]
            except KeyError:
                extra.append(key)
        result = []
        if different:
            different.sort()
            result.append('Mismatching, key: (expected != actual):')
            for r in different:
                result.append(r)
        if missing:
            missing.sort()
            result.append('The following keys were missing from actual:')
            for r in missing:
                result.append(repr(r))
        if extra:
            extra.sort()
            result.append('The following extra keys were found in actual:')
            for r in extra:
                result.append(repr(r))
        return result

    def _check_cache(self, cache, tcache):
        if self.echo:
            print('cache:')
            pprint(self.da._v_cache[0])
            print('tcache:')
            pprint(self.da._v_cache[1])
            print()
        result = []
        r = self._check_mapping(cache, self.da._v_cache[0])
        if r:
            result.append('cache did not match expected:')
            result.extend(r)
        r = self._check_mapping(tcache, self.da._v_cache[1])
        if r:
            result.append('tcache did not match expected:')
            result.extend(r)
        if result:
            self.fail('\n\n' + '\n'.join(result))

    def test_bad_aquisition(self):
        # checks that something called _v_cache isn't acquired from anywhere
        from ExtensionClass import Base

        class Dummy(Base):
            _v_cache = 'muhahaha'

        obj = Dummy()
        self.da = self.da.__of__(obj)
        del self.da._v_cache
        self._do_query('query', 1)

    def test_same_query_different_seconds(self):
        # this tests a sequence of requests for the same
        # query, but where the item returned is always in the cache
        self._check_cache({}, {})
        for t in range(1, 6):
            self._do_query('query', t)
            self._check_cache(
                {('query', 1, 'conn_id'): (1, 'result for query')},
                {1: ('query', 1, 'conn_id')})

    def test_same_query_same_second(self):
        # this tests a sequence set of requests for the same
        # query, but where the item returned is always in the cache
        # and where the queries all occur in the same second
        self._check_cache({}, {})
        for t in range(11, 16, 1):
            t = float(t) / 10
            self._do_query('query', t)
            self._check_cache(
                {('query', 1, 'conn_id'): (1.1, 'result for query')},
                {1.1: ('query', 1, 'conn_id')})

    def test_different_queries_different_second(self):
        # This tests different queries being fired into the cache
        # in sufficient volume to excercise the purging code
        self._check_cache({}, {})
        # one
        self._do_query('query1', 1.1)
        self._check_cache(
            {('query1', 1, 'conn_id'): (1.1, 'result for query1')},
            {1.1: ('query1', 1, 'conn_id')})
        # two
        self._do_query('query2', 3.2)
        self._check_cache(
            {('query1', 1, 'conn_id'): (1.1, 'result for query1'),
             ('query2', 1, 'conn_id'): (3.2, 'result for query2')},
            {1.1: ('query1', 1, 'conn_id'),
             3.2: ('query2', 1, 'conn_id')})
        # three - now we drop our first cache entry
        self._do_query('query3', 4.3)
        self._check_cache(
            {('query2', 1, 'conn_id'): (3.2, 'result for query2'),
             ('query3', 1, 'conn_id'): (4.3, 'result for query3')},
            {3.2: ('query2', 1, 'conn_id'),
             4.3: ('query3', 1, 'conn_id')})
        # four - now we drop our second cache entry
        self._do_query('query4', 8.4)
        self._check_cache(
            {('query3', 1, 'conn_id'): (4.3, 'result for query3'),
             ('query4', 1, 'conn_id'): (8.4, 'result for query4')},
            {4.3: ('query3', 1, 'conn_id'),
             8.4: ('query4', 1, 'conn_id')})

    def test_different_queries_same_second(self):
        # This tests different queries being fired into the cache
        # in the same second in sufficient quantities to exercise
        # the purging code
        self._check_cache({}, {})
        # one
        self._do_query('query1', 1.0)
        self._check_cache(
            {('query1', 1, 'conn_id'): (1.0, 'result for query1')},
            {1.0: ('query1', 1, 'conn_id')})
        # two
        self._do_query('query2', 1.1)
        self._check_cache(
            {('query1', 1, 'conn_id'): (1.0, 'result for query1'),
             ('query2', 1, 'conn_id'): (1.1, 'result for query2')},
            {1.0: ('query1', 1, 'conn_id'),
             1.1: ('query2', 1, 'conn_id')})
        # three - now we drop our first cache entry
        self._do_query('query3', 1.2)
        self._check_cache(
            {('query2', 1, 'conn_id'): (1.1, 'result for query2'),
             ('query3', 1, 'conn_id'): (1.2, 'result for query3')},
            {1.1: ('query2', 1, 'conn_id'),
             1.2: ('query3', 1, 'conn_id')})
        # four - now we drop another cache entry
        self._do_query('query4', 1.3)
        self._check_cache(
            {('query3', 1, 'conn_id'): (1.2, 'result for query3'),
             ('query4', 1, 'conn_id'): (1.3, 'result for query4')},
            {1.2: ('query3', 1, 'conn_id'),
             1.3: ('query4', 1, 'conn_id')})

    def test_different_queries_same_timestamp(self):
        # This tests different queries being fired into the cache with the
        # same timestamp. This expecially might happen on windows where the
        # resolution of time.time() is not that good. This tests a bug which
        # let to a KeyError when purging the cache.
        self._check_cache({}, {})
        # Two queries are added at the same time.time():
        self._do_query('query1', 1.0)
        self._do_query('query2', 1.0)
        # There is only one entry in the timestamp cache:
        self._check_cache(
            {('query1', 1, 'conn_id'): (1.0, 'result for query1'),
             ('query2', 1, 'conn_id'): (1.0, 'result for query2')},
            {1.0: ('query2', 1, 'conn_id')})
        # Querying first query after cache time out purges the second one
        # from the timestamp cache:
        self._do_query('query1', 11.0)
        self._check_cache(
            {('query1', 1, 'conn_id'): (11.0, 'result for query1'),
             ('query2', 1, 'conn_id'): (1.0, 'result for query2')},
            {11.0: ('query1', 1, 'conn_id')})
        # Querying the second query after cache time out let to the KeyError
        # while trying to delete the entry from the timestamp cache:
        self._do_query('query2', 12.0)
        self._check_cache(
            {('query1', 1, 'conn_id'): (11.0, 'result for query1'),
             ('query2', 1, 'conn_id'): (12.0, 'result for query2')},
            {11.0: ('query1', 1, 'conn_id'), 12.0: ('query2', 1, 'conn_id')})

    def test_time_tcache_expires(self):
        # This tests that once the cache purging code is triggered,
        # it will actively hoover out all expired cache entries

        # the first query gets cached
        self._do_query('query1', 1)
        self._check_cache(
            {('query1', 1, 'conn_id'): (1, 'result for query1')},
            {1: ('query1', 1, 'conn_id')})
        # the 2nd gets cached, the cache is still smaller than max_cache
        self._do_query('query2', 2)
        self._check_cache(
            {('query1', 1, 'conn_id'): (1, 'result for query1'),
             ('query2', 1, 'conn_id'): (2, 'result for query2')},
            {1: ('query1', 1, 'conn_id'),
             2: ('query2', 1, 'conn_id')})
        # the 3rd trips the max_cache trigger,  so both our old queries get
        # dumped because they are past their expiration time
        self._do_query('query', 23)
        self._check_cache(
            {('query', 1, 'conn_id'): (23, 'result for query')},
            {23: ('query', 1, 'conn_id')})

    def test_time_refreshed_cache(self):
        # This tests that when a cached query is expired when it comes
        # to check for a cached entry for that query, the stale entry is
        # removed and replaced with a fresh entry.

        # the first query gets cached
        self._do_query('query1', 1)
        self._check_cache(
            {('query1', 1, 'conn_id'): (1, 'result for query1')},
            {1: ('query1', 1, 'conn_id')})
        # do the same query much later, so new one gets cached
        self._do_query('query1', 12)
        self._check_cache(
            {('query1', 1, 'conn_id'): (12, 'result for query1')},
            {12: ('query1', 1, 'conn_id')})


class DummyDA:

    def __call__(self):
        conn = DummyDB()
        conn.result = ((), ())
        return conn

    sql_quote__ = 'I do not know what this is.'


class Hook:

    conn_num = 1

    def __call__(self):
        conn_to_use = 'conn' + str(self.conn_num)
        self.conn_num += 1
        return conn_to_use


class TestCacheKeys(TestCase):
    # These tests check that the keys used for caching are unique
    # in the right ways.

    def _cached_result(self, DB__, query, row_count, conn_id):
        self.cache_key = query, row_count, conn_id
        # we return something that can be safely turned into an empty Result
        return ((), ())

    def setUp(self):
        from Shared.DC.ZRDB.DA import DA
        self.da = DA('da', 'title', 'conn_id', 'arg1 arg2', 'some sql')
        self.da._cached_result = self._cached_result
        self.da.conn_id = DummyDA()
        # These need to be set so DA.__call__ tries for a cached result
        self.da.cache_time_ = 1
        self.da.max_cache_ = 1

    def test_default(self):
        self.da()
        self.assertEqual(self.cache_key, ('some sql', 1000, 'conn_id'))

    def test_different_max_rows(self):
        self.da.max_rows_ = 123
        self.da()
        self.assertEqual(self.cache_key, ('some sql', 123, 'conn_id'))

    def test_connection_hook(self):
        self.da.connection_hook = 'hook_method'
        self.da.hook_method = Hook()
        self.da.conn1 = DummyDA()
        self.da()
        self.assertEqual(self.cache_key, ('some sql', 1000, 'conn1'))
        self.da.conn2 = DummyDA()
        self.da()
        self.assertEqual(self.cache_key, ('some sql', 1000, 'conn2'))


class TestFullChain(TestCase):
    # This exercises both DA.__call__ and DA._cached_result.

    def setUp(self):
        from Shared.DC.ZRDB.DA import DA
        self.da = DA('da', 'title', 'conn_id', 'arg1 arg2', 'some sql')
        self.da.conn_id = DummyDA()

    def test_args_match(self):
        # This checks is that DA._cached_result's call signature
        # matches that expected by DA.__call__

        # These need to be set so DA.__call__ tries for a cached result
        self.da.cache_time_ = 1
        self.da.max_cache_ = 1
        # the actual test, will throw exceptions if things aren't right
        self.da()

    def test_cached_result_not_called_for_no_caching(self):
        # blow up the _cached_result method on our
        # test instance
        self.da._cached_result = None
        # check we never get there with the default "no cachine"
        self.da()
        # turn caching on
        self.da.cache_time_ = 1
        self.da.max_cache_ = 1
        # check that we get an exception
        self.assertRaises(TypeError, self.da)


def test_suite():
    suite = TestSuite()
    suite.addTest(makeSuite(TestCaching))
    suite.addTest(makeSuite(TestCacheKeys))
    suite.addTest(makeSuite(TestFullChain))
    return suite



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from setuptools import find_packages
from setuptools import setup


version = '3.7'

with open('README.rst') as f:
    README = f.read()

with open('CHANGES.rst') as f:
    CHANGES = f.read()

setup(
    name='Products.ZSQLMethods',
    version=version,
    url='https://github.com/zopefoundation/Products.ZSQLMethods',
    project_urls={
        'Issue Tracker': ('https://github.com/zopefoundation'
                          '/Products.ZSQLMethods/issues'),
        'Sources': 'https://github.com/zopefoundation/Products.ZSQLMethods',
    },
    license='ZPL 2.1',
    description='SQL method support for Zope.',
    author='Zope Foundation and Contributors',
    author_email='zope-dev@zope.org',
    long_description=(README + '\n' + CHANGES),
    packages=find_packages('src'),
    namespace_packages=['Products', 'Shared', 'Shared.DC'],
    package_dir={'': 'src'},
    classifiers=[
        'Development Status :: 6 - Mature',
        'Environment :: Web Environment',
        'Framework :: Zope :: 4',
        'License :: OSI Approved :: Zope Public License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Database',
        'Topic :: Database :: Front-Ends',
    ],
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    install_requires=[
        'setuptools',
        'six',
        'Zope >= 4.2.1',
        'Missing',
        'Record',
        'DocumentTemplate >= 3.2.2',
    ],
    include_package_data=True,
    zip_safe=False,
)



from unittest import TestCase
from unittest import TestSuite
from unittest import makeSuite

from six import StringIO

from ExtensionClass import Base
from Shared.DC.ZRDB import RDB
from Shared.DC.ZRDB.Results import Results


class Brain(Base):
    def __init__(self, *args):
        pass


Parent = Base()


class TestResults(TestCase):

    def test_results(self):
        r = Results(([{'name': 'foo', 'type': 'integer'},
                      {'name': 'bar', 'type': 'integer'}],
                     ((1, 2), (3, 4))),
                    brains=Brain,
                    parent=Parent)
        self.assertEqual(len(r), 2)
        row = r[0]
        self.assertEqual(row[0], 1)
        self.assertEqual(row[1], 2)
        self.assertEqual(row.foo, 1)
        self.assertEqual(row.bar, 2)
        self.assertEqual(row.FOO, 1)
        self.assertEqual(row.BAR, 2)
        row = r[1]
        self.assertEqual(row[0], 3)
        self.assertEqual(row[1], 4)
        self.assertEqual(row.foo, 3)
        self.assertEqual(row.bar, 4)
        self.assertEqual(row.FOO, 3)
        self.assertEqual(row.BAR, 4)
        self.assertTrue(isinstance(row, Brain))

    def test_rdb_file(self):
        infile = StringIO("""\
        foo\tbar
        2i\t2i
        1\t2
        3\t4\
        """)
        r = RDB.File(infile,
                     brains=Brain,
                     parent=Parent)
        self.assertEqual(len(r), 2)
        row = r[0]
        self.assertEqual(row[0], 1)
        self.assertEqual(row[1], 2)
        self.assertEqual(row.foo, 1)
        self.assertEqual(row.bar, 2)
        self.assertEqual(row.FOO, 1)
        self.assertEqual(row.BAR, 2)
        row = r[1]
        self.assertEqual(row[0], 3)
        self.assertEqual(row[1], 4)
        self.assertEqual(row.foo, 3)
        self.assertEqual(row.bar, 4)
        self.assertEqual(row.FOO, 3)
        self.assertEqual(row.BAR, 4)
        self.assertTrue(isinstance(row, Brain))


def test_suite():
    return TestSuite((makeSuite(TestResults),))



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import unittest

from Testing.makerequest import makerequest


class TestTM(unittest.TestCase):

    def _getTargetClass(self):
        from ..DA import DA
        return DA

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_instantiation(self):
        da = self._makeOne('test_id', 'Test Title', 'conn_id',
                           'foo bar', '<dtml-var bar>')
        self.assertEqual(da.getId(), 'test_id')
        self.assertEqual(da.title, 'Test Title')
        self.assertEqual(da.connection_id, 'conn_id')
        self.assertEqual(da.arguments_src, 'foo bar')
        self.assertEqual(da.src, '<dtml-var bar>')

    def test_manage_edit(self):
        da = self._makeOne('test_id', 'Test Title', 'conn_id',
                           'foo bar', '<dtml-var bar>')
        da.manage_edit('New Title', 'conn_2', 'bar baz', '<dtml-var baz>')
        self.assertEqual(da.getId(), 'test_id')
        self.assertEqual(da.title, 'New Title')
        self.assertEqual(da.connection_id, 'conn_2')
        self.assertEqual(da.arguments_src, 'bar baz')
        self.assertEqual(da.src, '<dtml-var baz>')

    def test_manage_advanced(self):
        klass = self._getTargetClass()
        da = self._makeOne('test_id', 'Test Title', 'conn_id',
                           'foo bar', '<dtml-var bar>')

        # Check defaults
        self.assertEqual(da.max_rows_, klass.max_rows_)
        self.assertEqual(da.max_cache_, klass.max_cache_)
        self.assertEqual(da.cache_time_, klass.cache_time_)
        self.assertEqual(da.class_name_, klass.class_name_)
        self.assertEqual(da.class_file_, klass.class_file_)
        self.assertEqual(da.allow_simple_one_argument_traversal,
                         klass.allow_simple_one_argument_traversal)
        self.assertEqual(da.template_class, klass.template_class)
        self.assertEqual(da.connection_hook, klass.connection_hook)

        da.manage_advanced(5, 50, 10, 'MyRecord', 'ZSQLMethods.TestRecord',
                           direct=True, connection_hook='foo')
        self.assertEqual(da.max_rows_, 5)
        self.assertEqual(da.max_cache_, 50)
        self.assertEqual(da.cache_time_, 10)
        self.assertEqual(da.class_name_, 'MyRecord')
        self.assertEqual(da.class_file_, 'ZSQLMethods.TestRecord')
        self.assertEqual(da.allow_simple_one_argument_traversal, True)
        self.assertEqual(da.template_class, klass.template_class)
        self.assertEqual(da.connection_hook, 'foo')

    def test_manage_DAVget(self):
        da = makerequest(self._makeOne('test_id', 'Test Title', 'conn_id',
                                       'foo bar', '<dtml-var bar>'))
        self.assertEqual(da.manage_DAVget(), DEFAULT_DAV_SOURCE)

        da.manage_edit('New Title', 'conn_2', 'bar baz', '<dtml-var baz>')
        da.manage_advanced(5, 50, 10, 'MyRecord', 'ZSQLMethods.TestRecord',
                           direct=True, connection_hook='foo')
        self.assertEqual(da.manage_DAVget(), CHANGED_DAV_SOURCE)

    def test_PUT(self):
        klass = self._getTargetClass()
        da = makerequest(self._makeOne('test_id', 'Test Title', 'conn_id',
                                       'foo bar', '<dtml-var bar>'))

        da.REQUEST.set('BODY', CHANGED_DAV_SOURCE)
        da.PUT(da.REQUEST, da.REQUEST.RESPONSE)
        self.assertEqual(da.getId(), 'test_id')
        self.assertEqual(da.title, 'New Title')
        self.assertEqual(da.connection_id, 'conn_2')
        self.assertEqual(da.arguments_src, 'bar baz')
        self.assertEqual(da.src, '<dtml-var baz>\n')
        self.assertEqual(da.max_rows_, 5)
        self.assertEqual(da.max_cache_, 50)
        self.assertEqual(da.cache_time_, 10)
        self.assertEqual(da.class_name_, 'MyRecord')
        self.assertEqual(da.class_file_, 'ZSQLMethods.TestRecord')
        self.assertEqual(da.allow_simple_one_argument_traversal, True)
        self.assertEqual(da.template_class, klass.template_class)
        self.assertEqual(da.connection_hook, 'foo')


DEFAULT_DAV_SOURCE = """\
<dtml-comment>
# vi:syntax=sql
title : Test Title
connection id : conn_id
arguments : foo bar
max_rows : 1000
max_cache : 100
cache_time : 0
class_name : 
class_file : 
connection_hook : 
allow_simple_one_argument_traversal : 
</dtml-comment>
<dtml-var bar>
"""  # NOQA: W291

CHANGED_DAV_SOURCE = """\
<dtml-comment>
# vi:syntax=sql
title : New Title
connection id : conn_2
arguments : bar baz
max_rows : 5
max_cache : 50
cache_time : 10
class_name : MyRecord
class_file : ZSQLMethods.TestRecord
connection_hook : foo
allow_simple_one_argument_traversal : True
</dtml-comment>
<dtml-var baz>
"""
